@@ -30,65 +30,38 @@ static inline void update_latest_addr(struct wireguard_peer *peer, struct sk_buf
 		socket_set_peer_endpoint(peer, &endpoint);
 }
 
-static inline int skb_prepare_header(struct sk_buff *skb)
+static inline int skb_prepare_header(struct sk_buff *skb, struct wireguard_device *wg)
 {
 	struct udphdr *udp;
 	size_t data_offset, data_len;
 	enum message_type message_type;
-
-	if (unlikely(skb->len < sizeof(struct iphdr)))
-		return -EINVAL;
-	if (unlikely(ip_hdr(skb)->version != 4 && ip_hdr(skb)->version != 6))
-		return -EINVAL;
-	if (unlikely(ip_hdr(skb)->version == 6 && skb->len < sizeof(struct ipv6hdr)))
-		return -EINVAL;
-
+	if (unlikely(skb_examine_untrusted_ip_hdr(skb) != skb->protocol || skb_transport_header(skb) < skb->head || (skb_transport_header(skb) + sizeof(struct udphdr)) >= skb_tail_pointer(skb)))
+		return -EINVAL; /* Bogus IP header */
 	udp = udp_hdr(skb);
 	data_offset = (u8 *)udp - skb->data;
-	if (unlikely(data_offset > U16_MAX)) {
-		net_dbg_skb_ratelimited("Packet has offset at impossible location from %pISpfsc\n", skb);
-		return -EINVAL;
-	}
-	if (unlikely(data_offset + sizeof(struct udphdr) > skb->len)) {
-		net_dbg_skb_ratelimited("Packet isn't big enough to have UDP fields from %pISpfsc\n", skb);
-		return -EINVAL;
-	}
+	if (unlikely(data_offset > U16_MAX || data_offset + sizeof(struct udphdr) > skb->len))
+		return -EINVAL;  /* Packet has offset at impossible location or isn't big enough to have UDP fields*/
 	data_len = ntohs(udp->len);
-	if (unlikely(data_len < sizeof(struct udphdr))) {
-		net_dbg_skb_ratelimited("UDP packet is reporting too small of a size from %pISpfsc\n", skb);
-		return -EINVAL;
-	}
-	if (unlikely(data_len > skb->len - data_offset)) {
-		net_dbg_skb_ratelimited("UDP packet is lying about its size from %pISpfsc\n", skb);
-		return -EINVAL;
-	}
+	if (unlikely(data_len < sizeof(struct udphdr) || data_len > skb->len - data_offset))
+		return -EINVAL;  /* UDP packet is reporting too small of a size or lying about its size */
 	data_len -= sizeof(struct udphdr);
 	data_offset = (u8 *)udp + sizeof(struct udphdr) - skb->data;
-	if (unlikely(!pskb_may_pull(skb, data_offset + sizeof(struct message_header)))) {
-		net_dbg_skb_ratelimited("Could not pull header into data section from %pISpfsc\n", skb);
+	if (unlikely(!pskb_may_pull(skb, data_offset + sizeof(struct message_header)) || pskb_trim(skb, data_len + data_offset) < 0))
 		return -EINVAL;
-	}
-	if (pskb_trim(skb, data_len + data_offset) < 0) {
-		net_dbg_skb_ratelimited("Could not trim packet from %pISpfsc\n", skb);
-		return -EINVAL;
-	}
 	skb_pull(skb, data_offset);
-	if (unlikely(skb->len != data_len)) {
-		net_dbg_skb_ratelimited("Final len does not agree with calculated len from %pISpfsc\n", skb);
-		return -EINVAL;
-	}
+	if (unlikely(skb->len != data_len))
+		return -EINVAL; /* Final len does not agree with calculated len */
 	message_type = message_determine_type(skb);
 	__skb_push(skb, data_offset);
-	if (unlikely(!pskb_may_pull(skb, data_offset + message_header_sizes[message_type]))) {
-		net_dbg_skb_ratelimited("Could not pull full header into data section from %pISpfsc\n", skb);
+	if (unlikely(!pskb_may_pull(skb, data_offset + message_header_sizes[message_type])))
 		return -EINVAL;
-	}
 	__skb_pull(skb, data_offset);
 	return message_type;
 }
 
 static void receive_handshake_packet(struct wireguard_device *wg, struct sk_buff *skb)
 {
+	static unsigned long last_under_load = 0; /* Yes this is global, so that our load calculation applies to the whole system. */
 	struct wireguard_peer *peer = NULL;
 	enum message_type message_type;
 	bool under_load;
@@ -98,19 +71,23 @@ static void receive_handshake_packet(struct wireguard_device *wg, struct sk_buff
 	message_type = message_determine_type(skb);
 
 	if (message_type == MESSAGE_HANDSHAKE_COOKIE) {
-		net_dbg_skb_ratelimited("Receiving cookie response from %pISpfsc\n", skb);
+		net_dbg_skb_ratelimited("%s: Receiving cookie response from %pISpfsc\n", netdev_pub(wg)->name, skb);
 		cookie_message_consume((struct message_handshake_cookie *)skb->data, wg);
 		return;
 	}
 
-	under_load = skb_queue_len(&wg->incoming_handshakes) >= MAX_QUEUED_INCOMING_HANDSHAKES / 2;
+	under_load = skb_queue_len(&wg->incoming_handshakes) >= MAX_QUEUED_INCOMING_HANDSHAKES / 8;
+	if (under_load)
+		last_under_load = jiffies;
+	else
+		under_load = time_is_after_jiffies(last_under_load + HZ);
 	mac_state = cookie_validate_packet(&wg->cookie_checker, skb, under_load);
 	if ((under_load && mac_state == VALID_MAC_WITH_COOKIE) || (!under_load && mac_state == VALID_MAC_BUT_NO_COOKIE))
 		packet_needs_cookie = false;
 	else if (under_load && mac_state == VALID_MAC_BUT_NO_COOKIE)
 		packet_needs_cookie = true;
 	else {
-		net_dbg_skb_ratelimited("Invalid MAC of handshake, dropping packet from %pISpfsc\n", skb);
+		net_dbg_skb_ratelimited("%s: Invalid MAC of handshake, dropping packet from %pISpfsc\n", netdev_pub(wg)->name, skb);
 		return;
 	}
 
@@ -123,11 +100,11 @@ static void receive_handshake_packet(struct wireguard_device *wg, struct sk_buff
 		}
 		peer = noise_handshake_consume_initiation(message, wg);
 		if (unlikely(!peer)) {
-			net_dbg_skb_ratelimited("Invalid handshake initiation from %pISpfsc\n", skb);
+			net_dbg_skb_ratelimited("%s: Invalid handshake initiation from %pISpfsc\n", netdev_pub(wg)->name, skb);
 			return;
 		}
 		update_latest_addr(peer, skb);
-		net_dbg_ratelimited("Receiving handshake initiation from peer %Lu (%pISpfsc)\n", peer->internal_id, &peer->endpoint.addr);
+		net_dbg_ratelimited("%s: Receiving handshake initiation from peer %Lu (%pISpfsc)\n", netdev_pub(wg)->name, peer->internal_id, &peer->endpoint.addr);
 		packet_send_handshake_response(peer);
 		break;
 	}
@@ -139,11 +116,11 @@ static void receive_handshake_packet(struct wireguard_device *wg, struct sk_buff
 		}
 		peer = noise_handshake_consume_response(message, wg);
 		if (unlikely(!peer)) {
-			net_dbg_skb_ratelimited("Invalid handshake response from %pISpfsc\n", skb);
+			net_dbg_skb_ratelimited("%s: Invalid handshake response from %pISpfsc\n", netdev_pub(wg)->name, skb);
 			return;
 		}
 		update_latest_addr(peer, skb);
-		net_dbg_ratelimited("Receiving handshake response from peer %Lu (%pISpfsc)\n", peer->internal_id, &peer->endpoint.addr);
+		net_dbg_ratelimited("%s: Receiving handshake response from peer %Lu (%pISpfsc)\n", netdev_pub(wg)->name, peer->internal_id, &peer->endpoint.addr);
 		if (noise_handshake_begin_session(&peer->handshake, &peer->keypairs, true)) {
 			timers_ephemeral_key_created(peer);
 			timers_handshake_complete(peer);
@@ -171,17 +148,13 @@ static void receive_handshake_packet(struct wireguard_device *wg, struct sk_buff
 
 void packet_process_queued_handshake_packets(struct work_struct *work)
 {
-	struct wireguard_device *wg = container_of(work, struct wireguard_device, incoming_handshakes_work);
+	struct wireguard_device *wg = container_of(work, struct handshake_worker, work)->wg;
 	struct sk_buff *skb;
-	size_t num_processed = 0;
 
 	while ((skb = skb_dequeue(&wg->incoming_handshakes)) != NULL) {
 		receive_handshake_packet(wg, skb);
 		dev_kfree_skb(skb);
-		if (++num_processed == MAX_BURST_INCOMING_HANDSHAKES) {
-			queue_work(wg->handshake_wq, &wg->incoming_handshakes_work);
-			return;
-		}
+		cond_resched();
 	}
 }
 
@@ -201,7 +174,7 @@ static void keep_key_fresh(struct wireguard_peer *peer)
 
 	if (send) {
 		peer->sent_lastminute_handshake = true;
-		packet_queue_handshake_initiation(peer);
+		packet_queue_handshake_initiation(peer, false);
 	}
 }
 
@@ -210,6 +183,7 @@ void packet_consume_data_done(struct sk_buff *skb, struct wireguard_peer *peer,
 	struct net_device *dev;
 	struct wireguard_peer *routed_peer;
 	struct wireguard_device *wg;
+	unsigned int len;
 
 	socket_set_peer_endpoint(peer, endpoint);
 
@@ -219,40 +193,50 @@ void packet_consume_data_done(struct sk_buff *skb, struct wireguard_peer *peer,
 	if (unlikely(used_new_key)) {
 		peer->sent_lastminute_handshake = false;
 		packet_send_queue(peer);
+		timers_handshake_complete(peer);
 	}
 
 	keep_key_fresh(peer);
 
 	/* A packet with length 0 is a keepalive packet */
 	if (unlikely(!skb->len)) {
-		net_dbg_ratelimited("Receiving keepalive packet from peer %Lu (%pISpfsc)\n", peer->internal_id, &peer->endpoint.addr);
-		goto packet_processed;
-	}
-
-	if (!pskb_may_pull(skb, 1 /* For checking the ip version below */)) {
-		++dev->stats.rx_errors;
-		++dev->stats.rx_length_errors;
-		net_dbg_ratelimited("Packet missing IP version from peer %Lu (%pISpfsc)\n", peer->internal_id, &peer->endpoint.addr);
+		net_dbg_ratelimited("%s: Receiving keepalive packet from peer %Lu (%pISpfsc)\n", netdev_pub(peer->device)->name, peer->internal_id, &peer->endpoint.addr);
 		goto packet_processed;
 	}
+	if (skb_network_header(skb) < skb->head)
+		goto dishonest_packet_size;
 
 	skb->dev = dev;
 	skb->ip_summed = CHECKSUM_UNNECESSARY;
-	if (skb->len >= sizeof(struct iphdr) && ip_hdr(skb)->version == 4) {
-		skb->protocol = htons(ETH_P_IP);
+	skb->protocol = skb_examine_untrusted_ip_hdr(skb);
+	if (skb->protocol == htons(ETH_P_IP)) {
+		len = ntohs(ip_hdr(skb)->tot_len);
+		if (unlikely(len < sizeof(struct iphdr)))
+			goto dishonest_packet_size;
 		if (INET_ECN_is_ce(PACKET_CB(skb)->ds))
 			IP_ECN_set_ce(ip_hdr(skb));
-	} else if (skb->len >= sizeof(struct ipv6hdr) && ip_hdr(skb)->version == 6) {
-		skb->protocol = htons(ETH_P_IPV6);
+
+	} else if (skb->protocol == htons(ETH_P_IPV6)) {
+		len = ntohs(ipv6_hdr(skb)->payload_len) + sizeof(struct ipv6hdr);
 		if (INET_ECN_is_ce(PACKET_CB(skb)->ds))
 			IP6_ECN_set_ce(skb, ipv6_hdr(skb));
 	} else {
 		++dev->stats.rx_errors;
-		++dev->stats.rx_length_errors;
-		net_dbg_ratelimited("Packet neither ipv4 nor ipv6 from peer %Lu (%pISpfsc)\n", peer->internal_id, &peer->endpoint.addr);
+		++dev->stats.rx_frame_errors;
+		net_dbg_ratelimited("%s: Packet neither ipv4 nor ipv6 from peer %Lu (%pISpfsc)\n", netdev_pub(peer->device)->name, peer->internal_id, &peer->endpoint.addr);
 		goto packet_processed;
 	}
 
+	if (unlikely(len > skb->len)) {
+dishonest_packet_size:
+		net_dbg_ratelimited("%s: Packet is lying about its size from peer %Lu (%pISpfsc)\n", netdev_pub(peer->device)->name, peer->internal_id, &peer->endpoint.addr);
+		++dev->stats.rx_errors;
+		++dev->stats.rx_length_errors;
+		goto packet_processed;
+	}
+	if (len < skb->len && pskb_trim(skb, len))
+		goto packet_processed;
+
 	timers_data_received(peer);
 
 	routed_peer = routing_table_lookup_src(&wg->peer_routing_table, skb);
@@ -261,15 +245,16 @@ void packet_consume_data_done(struct sk_buff *skb, struct wireguard_peer *peer,
 	if (unlikely(routed_peer != peer)) {
 		++dev->stats.rx_errors;
 		++dev->stats.rx_frame_errors;
-		net_dbg_skb_ratelimited("Packet has unallowed src IP (%pISc) from peer %Lu (%pISpfsc)\n", skb, peer->internal_id, &peer->endpoint.addr);
+		net_dbg_skb_ratelimited("%s: Packet has unallowed src IP (%pISc) from peer %Lu (%pISpfsc)\n", netdev_pub(peer->device)->name, skb, peer->internal_id, &peer->endpoint.addr);
 		goto packet_processed;
 	}
 
+	len = skb->len;
 	if (likely(netif_rx(skb) == NET_RX_SUCCESS))
-		rx_stats(peer, skb->len);
+		rx_stats(peer, len);
 	else {
 		++dev->stats.rx_dropped;
-		net_dbg_ratelimited("Failed to give packet to userspace from peer %Lu (%pISpfsc)\n", peer->internal_id, &peer->endpoint.addr);
+		net_dbg_ratelimited("%s: Failed to give packet to userspace from peer %Lu (%pISpfsc)\n", netdev_pub(peer->device)->name, peer->internal_id, &peer->endpoint.addr);
 	}
 	goto continue_processing;
 
@@ -283,27 +268,34 @@ continue_processing:
 
 void packet_receive(struct wireguard_device *wg, struct sk_buff *skb)
 {
-	int message_type = skb_prepare_header(skb);
+	int message_type = skb_prepare_header(skb, wg);
 	if (unlikely(message_type < 0))
 		goto err;
 	switch (message_type) {
 	case MESSAGE_HANDSHAKE_INITIATION:
 	case MESSAGE_HANDSHAKE_RESPONSE:
-	case MESSAGE_HANDSHAKE_COOKIE:
+	case MESSAGE_HANDSHAKE_COOKIE: {
+		int cpu_index, cpu, target_cpu;
 		if (skb_queue_len(&wg->incoming_handshakes) > MAX_QUEUED_INCOMING_HANDSHAKES) {
-			net_dbg_skb_ratelimited("Too many handshakes queued, dropping packet from %pISpfsc\n", skb);
+			net_dbg_skb_ratelimited("%s: Too many handshakes queued, dropping packet from %pISpfsc\n", netdev_pub(wg)->name, skb);
 			goto err;
 		}
 		skb_queue_tail(&wg->incoming_handshakes, skb);
+		/* Select the CPU in a round-robin */
+		cpu_index = ((unsigned int)atomic_inc_return(&wg->incoming_handshake_seqnr)) % cpumask_weight(cpu_online_mask);
+		target_cpu = cpumask_first(cpu_online_mask);
+		for (cpu = 0; cpu < cpu_index; ++cpu)
+			target_cpu = cpumask_next(target_cpu, cpu_online_mask);
 		/* Queues up a call to packet_process_queued_handshake_packets(skb): */
-		queue_work(wg->handshake_wq, &wg->incoming_handshakes_work);
+		queue_work_on(target_cpu, wg->incoming_handshake_wq, &per_cpu_ptr(wg->incoming_handshakes_worker, target_cpu)->work);
 		break;
+	}
 	case MESSAGE_DATA:
 		PACKET_CB(skb)->ds = ip_tunnel_get_dsfield(ip_hdr(skb), skb);
 		packet_consume_data(skb, wg);
 		break;
 	default:
-		net_dbg_skb_ratelimited("Invalid packet from %pISpfsc\n", skb);
+		net_dbg_skb_ratelimited("%s: Invalid packet from %pISpfsc\n", netdev_pub(wg)->name, skb);
 		goto err;
 	}
 	return;
